#!/usr/bin/env python

"""Caching API for Storage Connect plug-ins.

Some of the operations provided by storage servers are very expensive.
They can take seconds or more to complete.  Since they often need to be
repeated, it makes sense to cache responses for a limited period of
time.

Any object may serve as a cache as long as it provides the following API::

    set(ss_uuid, key, value, expire_seconds=30)
      - Stores a value for expire_seconds
    get(ss_uuid, key)
      - Returns the value for key, or None if key has expired or was
        never stored
    extend(ss_uuid, key, extend_seconds)
      - Extends the expire time for key; does nothing if key has already
        expired
    clear(ss_uuid)
      - Drops all keys for ss_uuid

"""


"""
@copyright: Copyright (C) 2010 Oracle and/or its affiliates. All rights reserved.

@license:   See the file COPYING for redistribution information.

@summary:   Caching API for Storage Connect plug-ins.
"""


import time


__all__ = ['OSCDefaultCache']


class OSCDefaultCache(object):

    """A basic in-memory cache for simple users

    This simple cache provides the basic caching API for via an
    in-memory store.  No data is persisted, thus it is not available
    to other processes.

    """

    def __init__(self):
        self._data = {}
        self._expire = {}

    def _check_expire(self, ss_uuid, key):
        ss_cache = self._data.get(ss_uuid, {})
        ss_expire = self._expire.get(ss_uuid, {})
        if ss_cache.has_key(key) and (time.time() > ss_expire[key]):
            del ss_cache[key]
            del ss_expire[key]

    def get(self, ss_uuid, key):
        """Retrieve a cached value"""

        self._check_expire(ss_uuid, key)
        ss_cache = self._data.get(ss_uuid, {})
        return ss_cache.get(key, None)

    def set(self, ss_uuid, key, value, expire_seconds=30):
        """Store a cached value for the specified server"""

        if not self._data.has_key(ss_uuid):
            self._data[ss_uuid] = {}
        self._data[ss_uuid][key] = value
        if not self._expire.has_key(ss_uuid):
            self._expire[ss_uuid] = {}
        self._expire[ss_uuid][key] = time.time() + expire_seconds

    def extend(self, ss_uuid, key, extend_seconds):
        """Extend the expire timeout for a cached value"""

        self._check_expire(ss_uuid, key)
        if self._expire.get(ss_uuid, {}).has_key(key):
            self._expire[ss_uuid][key] += extend_seconds

    def clear(self, ss_uuid):
        """Clear the cache for the specified server"""

        if self._data.has_key(ss_uuid):
            del self._data[ss_uuid]
        if self._expire.has_key(ss_uuid):
            del self._expire[ss_uuid]

    def read(self, ss_uuid):
        """read the entire cache for the specified server"""

        if not self._data.has_key(ss_uuid):
            return None
        return self._data[ss_uuid]
        
    def write(self, ss_uuid, data):
        """write the entire cache for the specified server"""

        if self._data.has_key(ss_uuid):
            self.clear(ss_uuid)
        self._data[ss_uuid] = data
