#!/usr/bin/env python

"""
@copyright: Copyright (C) 2008, 2010 Oracle and/or its affiliates. All rights reserved.

@license:   See the file COPYING for redistribution information.

@summary:   Miscellaneous utility functions that is available for use by any
            Storage Connect plug-in.
"""

import os
import os.path
import re
import fcntl
import socket
import struct
import gettext
import fileinput
import subprocess
import string
import time



from glob import glob


_ = gettext.gettext

def checkRequired(*required_commands):
    """
    Check if the command (full path to the command) is available on the system,
    if not it will raise the standardized Storage Connect exception.

    @param required_commands: Commands to check if available.
    @type  required_commands: C{str} or C{list}

    @raise CommandMissingEx: Raised if a command cannot be found.
    """
    from OSCPlugin import CommandMissingEx

    for required_command in required_commands:
        if type(required_command) == list:
            checkRequired(required_command)
        elif type(required_command) == str:
            if not os.path.exists(required_command):
                raise CommandMissingEx(_("Missing command: %s" %
                                         required_command))



def resizeMPDev(mp_dev):
    """
    This will resize a multipath device.

    @param mp_dev: The multipath device (dm device) to resize.
    @type  mp_dev: C{str}
    """

    from OSCPlugin import IPlugin
    from OSCPlugin import OperationFailedEx

    if not os.path.exists(IPlugin.dev_path_prefix):
        IPlugin.logger.info("resizeMPDev() - "
                            "Device prefix dir (%s) does not exist" %
                            IPlugin.dev_path_prefix)
        return


    if os.path.exists(mp_dev):
        mpl_p = subprocess.Popen(["multipath",
                                  "-l",
                                  os.path.basename(mp_dev)],
                                 stdout = subprocess.PIPE,
                                 stderr = subprocess.PIPE,
                                 close_fds=True)
        (p_out, p_err) = mpl_p.communicate()

        if mpl_p.returncode != 0:
            IPlugin.logger.error("resizeMPDev() - Unable to obtain "
                                 "multipath info for %s: %s",
                                 mp_dev, p_err)
            raise OperationFailedEx(_("Unable to obtain multipath info for "
                                      "%s: %s" % (mp_dev, p_err)))

        slaves = []
        for mpl_line in p_out.splitlines(): #pylint: disable=E1103
            mpl_fields = mpl_line.split()
            for index in range(len(mpl_fields)):
                if (mpl_fields[index] == "|-" or mpl_fields[index] == "`-"):
                    slaves.append(mpl_fields[index + 2])


        for slave in slaves:
            dev_name = os.path.join("/dev", slave)

            if os.path.exists(dev_name):
                IPlugin.logger.debug("resizeMPDev() - reading device %s partition table " % slave)

                dms_p = subprocess.Popen(["blockdev",
                                          "--rereadpt",
                                          dev_name],
                                          stdout = subprocess.PIPE,
                                          stderr = subprocess.PIPE,
                                          close_fds=True)
                (p_out, p_err) = dms_p.communicate()

                if dms_p.returncode != 0:
                    if p_err.find('Input/output error') < 0:
                        IPlugin.logger.error("Unable to read %s partition table: %s" % (dev_name, p_err))
                        raise OperationFailedEx(_("Unable to read %s partition table: %s" % (dev_name, p_err)))

                    
        cmd = "/sbin/multipathd -k'resize multipath %s' >/dev/null" % os.path.basename(mp_dev)

        try:
            os.system(cmd)
       
        except:
            IPlugin.logger.error("resizeMPDev() - Unable to resize %s", mp_dev)
            raise OperationFailedEx(_("Unable to resize %s" % mp_dev))

            
def _disable_dm_queue(mp_dev):
    """
    This will cause all multipath queued I/O to fail immediately

    @param mp_dev: The multipath device (dm device) that need to disable.
    @type  mp_dev: C{str}
    """

    from OSCPlugin import IPlugin

    IPlugin.logger.debug("_disable_dm_queue() - disable multipath queue setting %s" % mp_dev)
    dmm_p = subprocess.Popen(["dmsetup",
                              "message",
                              os.path.basename(mp_dev),
                              "0", "fail_if_no_path"],
                             stdout = subprocess.PIPE,
                             stderr = subprocess.PIPE,
                             close_fds=True)
    (p_out, p_err) = dmm_p.communicate()

    if dmm_p.returncode != 0:
        IPlugin.logger.error("Unable to disable multipath queue setting %s: %s" %
                             (mp_dev, p_err))

def _flush_dm_devmap(mp_dev):
    """
    This will flush a multipath device map

    @param mp_dev: The multipath device (dm device) that need to flush.
    @type  mp_dev: C{str}
    """

    from OSCPlugin import IPlugin

    IPlugin.logger.debug("_flush_dm_devmap() - flush multipath device map %s" % mp_dev)
    mul_p = subprocess.Popen(["multipath",
                              "-f",
                              os.path.basename(mp_dev)],
                             stdout = subprocess.PIPE,
                             stderr = subprocess.PIPE,
                             close_fds=True)
    (p_out, p_err) = mul_p.communicate()

    if mul_p.returncode != 0:
        IPlugin.logger.error("Unable to flush multipath device map %s: %s" %
                             (mp_dev, p_err))

def destroyMPDev(mp_dev):
    """
    This will destroy a multipath device as well as all of its children
    (partitions and block devices).

    @param mp_dev: The multipath device (dm device) that need to teared down.
    @type  mp_dev: C{str}
    """
    from OSCPlugin import IPlugin
    from OSCPlugin import OperationFailedEx


    if not os.path.exists(IPlugin.dev_path_prefix):
        IPlugin.logger.info("destroyMPDev() - "
                            "Device prefix dir (%s) does not exist" %
                            IPlugin.dev_path_prefix)
        return

    dev_names = os.listdir(IPlugin.dev_path_prefix)
    p_lst = []
    d_lst = [d_lst for d_lst in dev_names
                if re.match("%sp\d+" % os.path.basename(mp_dev), d_lst)]
    if len(d_lst) > 0:
        p_lst = p_lst + d_lst[:]

    for part in p_lst:
        if os.path.exists(os.path.join(IPlugin.dev_path_prefix, part)):
            IPlugin.logger.info("destroyMPDev() - Removing DM device %s" % part)
            dms_p = subprocess.Popen(["dmsetup",
                                      "remove",
                                      part],
                                     stdout = subprocess.PIPE,
                                     stderr = subprocess.PIPE,
                                     close_fds=True)
            (p_out, p_err) = dms_p.communicate()

            if dms_p.returncode != 0:
                IPlugin.logger.error("destroyMPDev() - Unable to tear down "
                                     "multipath device: %s" % p_err)
                raise OperationFailedEx(_("Unable to tear down multipath"
                                          "device: %s" % p_err))

    if os.path.exists(mp_dev):
        mpl_p = subprocess.Popen(["multipath",
                                  "-l",
                                  os.path.basename(mp_dev)],
                                 stdout = subprocess.PIPE,
                                 stderr = subprocess.PIPE,
                                 close_fds=True)
        (p_out, p_err) = mpl_p.communicate()

        if mpl_p.returncode != 0:
            IPlugin.logger.error("destroyMPDev() - Unable to obtain "
                                 "multipath info for %s: %s",
                                 mp_dev, p_err)
            raise OperationFailedEx(_("Unable to obtain multipath info for "
                                      "%s: %s" % (mp_dev, p_err)))

        slaves = []
        for mpl_line in p_out.splitlines(): #pylint: disable=E1103
            mpl_fields = mpl_line.split()
            for index in range(len(mpl_fields)):
                if (mpl_fields[index] == "|-" or mpl_fields[index] == "`-"):
                    slaves.append(mpl_fields[index + 2])

        IPlugin.logger.info("destroyMPDev() - Removing DM device %s" % mp_dev)
        dms_p = subprocess.Popen(["dmsetup",
                                  "remove",
                                  os.path.basename(mp_dev)],
                                 stdout = subprocess.PIPE,
                                 stderr = subprocess.PIPE,
                                 close_fds=True)
        (p_out, p_err) = dms_p.communicate()

        if dms_p.returncode != 0:
            IPlugin.logger.error("Unable to tear down multipath device %s: %s" %
                                 (mp_dev, p_err))

            _disable_dm_queue(mp_dev) # disable multipath queue setting

            time.sleep(2.5)  # wait for queue I/O to fail before flush device map

            _flush_dm_devmap(mp_dev) # flush multipath device map

            time.sleep(0.5)  # wait for flush multipath device map

            if os.path.exists(mp_dev):  # check for the DM device  
                IPlugin.logger.debug("destroyMPDev() - Removing DM device %s" % mp_dev)
                dmr_p = subprocess.Popen(["dmsetup",
                                          "remove",
                                          os.path.basename(mp_dev)],
                                         stdout = subprocess.PIPE,
                                         stderr = subprocess.PIPE,
                                         close_fds=True)
                (p_out, p_err) = dmr_p.communicate()

                if dmr_p.returncode != 0:
                    IPlugin.logger.error("Unable to tear down multipath device %s: %s" %
                                         (mp_dev, p_err))
                    raise OperationFailedEx(_("Unable to tear down multipath "
                                              "device %s: %s" % (mp_dev, p_err)))

                IPlugin.logger.debug("tear down multipath device %s: success" % mp_dev)

        for slave in slaves:
            devdel_name = os.path.join(os.path.join("/sys/block", slave),
                                       "device/delete")

            if os.path.exists(devdel_name):
                IPlugin.logger.debug("destroyMPDev() - Deleting block "
                                     "device %s" % slave)
                try:
                    devdel_f = open(devdel_name, "a")
                    devdel_f.write("1")
                    devdel_f.close()

                except IOError, ioe:
                    IPlugin.logger.error("IOError deleting device %s: %s",
                                         slave, ioe.strerror)

def resizePPDev(ss_record, pp_dev):
    """
    This will resize a powerpath device.

    @param pp_dev: The powerpath device to resize.
    @type  pp_dev: C{str}
    """

    from OSCPlugin import IPlugin
    from OSCPlugin import OperationFailedEx

    if not os.path.exists(IPlugin.dev_path_prefix):
        IPlugin.logger.info("resizePPDev() - "
                            "Device prefix dir (%s) does not exist" %
                            IPlugin.dev_path_prefix)
        return

    if ss_record["storage_type"] == IPlugin.iSCSIStorage:
            
        iscsi_p = subprocess.Popen([IPlugin.iscsiadm,
                                    "--mode", "node",
                                    "--rescan"],
                                   stdout = subprocess.PIPE,
                                   stderr = subprocess.PIPE,
                                   close_fds=True)
        (p_out, p_err) = iscsi_p.communicate()
        if iscsi_p.returncode != 0:
            IPlugin.logger.error("Target refresh failed: %s" % p_err)
            raise RefreshFailedEx(_("Target refresh failed: %s" % p_err))
    
        time.sleep(1.2)
        
        
    
    elif ss_record["storage_type"] == IPlugin.SANStorage:
        if os.path.exists("/sys/class/fc_host"):
            for fc_host in os.listdir("/sys/class/scsi_host"):
                try:
                    if os.path.exists("/sys/class/scsi_host/%s/scan" % fc_host):
                        scan_f = open("/sys/class/scsi_host/%s/scan" % fc_host,
                                  "w", 0)
                        scan_f.write("- - -")
                        scan_f.close()
                    else:
                        IPlugin.logger.info("/sys/class/scsi_host/%s/scan does not exist" % fc_host)


                except IOError, ioe:
                    IPlugin.logger.error("IO Error: %s" % ioe.strerror)
                    raise RefreshFailedEx(_("IO Error: %s" % ioe.strerror))
                

def destroyPPDev(pp_dev):
    """
    This will destroy a powerpath device as well as all of its children
    (partitions and block devices).

    @param pp_dev: The powerpath device (pp device) that need to teared down.
    @type  pp_dev: C{str}
    """
    from OSCPlugin import IPlugin
    from OSCPlugin import OperationFailedEx


    if not os.path.exists(IPlugin.dev_path_prefix):
        IPlugin.logger.info("destroyPPDev() - "
                            "Device prefix dir (%s) does not exist" %
                            IPlugin.dev_path_prefix)
        return


    if os.path.exists(pp_dev):

        link = os.path.basename(os.readlink(pp_dev))

        slaves = getPPSlaves(link)

        power_p = subprocess.Popen(["/sbin/powermt",
                                 "remove",
                                 "dev=%s" % link],
                                 stdout = subprocess.PIPE,
                                 stderr = subprocess.PIPE,
                                 close_fds=True)
        (p_out, p_err) = power_p.communicate()  

        if power_p.returncode != 0:
            IPlugin.logger.error("destroyPPDev() - Unable to remove "
                                 "powerpath device for %s: %s",
                                 pp_dev, p_err)
            raise OperationFailedEx(_("Unable to remove powerpath device for "
                                      "%s: %s" % (pp_dev, p_err)))
            return

        power_p = subprocess.Popen(["/sbin/powermt",
                             "release"],
                             stdout = subprocess.PIPE,
                             stderr = subprocess.PIPE,
                             close_fds=True)
        (p_out, p_err) = power_p.communicate()  
        if power_p.returncode != 0:
            IPlugin.logger.error("destroyPPDev() - Unable to release "
                                 "powerpath device for %s: %s",
                                 pp_dev, p_err)
            raise OperationFailedEx(_("Unable to release powerpath device for "
                                      "%s: %s" % (pp_dev, p_err)))
            return

        
        for slave in slaves:
            devdel_name = os.path.join(os.path.join("/sys/block", slave),
                                       "device/delete")

            if os.path.exists(devdel_name):
                IPlugin.logger.debug("destroyPPDev() - Deleting block "
                                     "device %s" % slave)
                try:
                    devdel_f = open(devdel_name, "a")
                    devdel_f.write("1")
                    devdel_f.close()

                except IOError, ioe:
                    IPlugin.logger.error("IOError deleting device %s: %s",
                                         slave, ioe.strerror)

def getPPSlaves(pp_dev):
    from OSCPlugin import IPlugin

    scsi_devs = []

    power_p = subprocess.Popen(["/sbin/powermt",
                                 "display",
                                 "dev=%s" % pp_dev],
                                 stdout = subprocess.PIPE,
                                 stderr = subprocess.PIPE,
                                 close_fds=True)
    (p_out, p_err) = power_p.communicate()  

    if power_p.returncode != 0:
        IPlugin.logger.error("getPPSlaves() - Unable to enumerate "
                             "scsi devices for %s: %s",
                             pp_dev, p_err)
        return scsi_devs

    power_lines_LIST = map( string.strip, p_out.splitlines() )

    entries = len(power_lines_LIST)

    if entries < 9:
        return scsi_devs


    for index in range(9, entries - 3):
        items = power_lines_LIST[index].split()
        sd = re.search(r' sd[a-z]*', power_lines_LIST[index])
        if sd:
            scsi_devs.append(sd.group().strip())
           

    return scsi_devs



def getPage83FromMPDev(mp_dev):
    """
    This will return the page83_id from the supplied multipath device.

    @param mp_dev: The multi path device (dm device) to interrogate.
    @type  mp_dev: C{str}
    """

    from OSCPlugin import IPlugin
    from OSCPlugin import OperationFailedEx

    if not os.path.exists(mp_dev):
        IPlugin.logger.info("getPage83FromMPDev() - Device (%s) does not exist" % mp_dev)
        return

    dms_p = subprocess.Popen(["dmsetup",
                              "info",
                              "--columns",
                              "--noheadings",
                              "--options=UUID",
                              os.path.basename(mp_dev)],
                              stdout = subprocess.PIPE,
                              stderr = subprocess.PIPE,
                              close_fds=True)
    (p_out, p_err) = dms_p.communicate()

    if dms_p.returncode != 0:
        IPlugin.logger.error("Unable to get uuid of the multipath device %s: %s" %
                             (mp_dev, p_err))
        raise OperationFailedEx(_("Unable to get uuid of the multipath device %s: %s" % 
                                  (mp_dev, p_err)))

    return p_out.rstrip("\n").replace("mpath-", "") #pylint: disable=E1103

def containsFile(fs_record, file_path):
    """
    Check if the file is contained on the file system identified by the
    File System record.

    @param fs_record: L{File System record<OSCPlugin.IPlugin.__FSRecord__>}
    @type  fs_record: C{dict}

    @param file_path: File path to check.
    @type  file_path: C{str}

    @return: True if the file is contained in the file system.
    @rtype:  C{bool}

    @raise IPluginException: on failure
    """
    if file_path.startswith('/'):
        if file_path[:len(fs_record["mount_path"])] == fs_record["mount_path"]:
            return True
        else:
            return False

    else:
        return True

def getFileRecord(fs_record, file_path):
    """
    Generate a file record for the file identified by the file path for
    the specific File System record.

    @param fs_record: L{File System record<OSCPlugin.IPlugin.__FSRecord__>}
    @type  fs_record: C{dict}

    @param file_path: File path to create the file record for.
    @type  file_path: C{str}

    @return: L{File record<OSCPlugin.IPlugin.__FileRecord__>}
    @rtype:  C{dict}

    @raise IPluginException: on failure
    """
    from OSCPlugin import ValueFormatEx

    if file_path.startswith('/'):
        if file_path[:len(fs_record["mount_path"])] != fs_record["mount_path"]:
            raise ValueFormatEx(_("File (%s) is not contained on %s file "
                                 "system" % (file_path, fs_record["name"])))

        else:
            if ".snap" in file_path:
                fr_type = IFileSystemPlugin.SnapType

            else:
                fr_type = IFileSystemPlugin.FileType

            return {"fr_type":      fr_type,
                    "file_path":    file_path}

    else:
        full_path = os.path.join(fs_record["mount_path"], file_path)
        if ".snap" in full_path:
            fr_type = IFileSystemPlugin.SnapType

        else:
            fr_type = IFileSystemPlugin.FileType

        return {"fr_type":      fr_type,
                "file_path":    file_path}

def iSCSIDiscoveryAuth(username, passwd, chap):
    """
    Generate a file record for the file identified by the file path for
    the specific File System record.

    @param ss_record: L{Storage Server record<OSCPlugin.IPlugin.__SSRecord__>}
    @type  ss_record: C{dict}

    @raise IPluginException: on failure
    """
    
    from OSCPlugin import TargetDiscoveryEx

    found_username = False
    found_passwd = False
    found_chap = False
    
  
    try:
        cnf_file = fileinput.input("/etc/iscsi/iscsid.conf",
                                   inplace=1,
                                   backup=".bak")
        for cnf_line in cnf_file:
            if username and re.match("discovery.sendtargets.auth.username.*", cnf_line):
                cnf_line = "discovery.sendtargets.auth.username = %s\n" % username
                found_username = True

            if passwd and re.match("discovery.sendtargets.auth.password.*", cnf_line):
                cnf_line = "discovery.sendtargets.auth.password = %s\n" % passwd
                found_passwd = True

            if chap and re.match("discovery.sendtargets.auth.authmethod", cnf_line):
                cnf_line = "discovery.sendtargets.auth.authmethod = CHAP\n"
                found_chap = True

            print cnf_line,

        cnf_file = open("/etc/iscsi/iscsid.conf", "a")
        if username and not found_username:
            print >> cnf_file, "discovery.sendtargets.auth.username = %s" % username

        if passwd and not found_passwd:
            print >> cnf_file, "discovery.sendtargets.auth.password = %s" % passwd

        if chap and not found_chap:
            print >> cnf_file, "discovery.sendtargets.auth.authmethod = CHAP"

    except Exception, exc:
        cnf_file.close()
        if os.path.exists("/etc/iscsi/iscsid.conf.bak"):
            os.remove("/etc/iscsi/iscsid.conf")
            os.rename("/etc/iscsi/iscsid.conf.bak",
                      "/etc/iscsi/iscsid.conf")

        raise TargetDiscoveryEx(_(exc))

    cnf_file.close()

def getNetDevInfo(netdev):
    """
    Obtain network device information.

    @param netdev: Network device name
    @type  netdev: C{str}

    @return: Network device information
    @rtype:  C{dict}
    """
    netdev_info = {"netdev": netdev}
    temp_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock_fd = temp_sock.fileno()

    try:
        netdev_info["addr"]   = __netdev_ioctl__(sock_fd, 0x8915, netdev)
        netdev_info["mask"]   = __netdev_ioctl__(sock_fd, 0x891b, netdev)
        netdev_info["bcast"]  = __netdev_ioctl__(sock_fd, 0x8919, netdev)
        netdev_info["hwaddr"] = __netdev_ioctl__(sock_fd, 0x8927, netdev)

    except:
        pass

    temp_sock.close()
    return netdev_info

def __netdev_ioctl__(sock_fd, ioctl_num, dev_name): #pylint: disable=C0103
    """
    Execute specific ioctl op for network devices.

    @param sock_fd: File descriptor
    @type  sock_fd: C{int}

    @param ioctl_num: ioctl operation
    @type  ioctl_num: C{int}

    @param dev_name: Device name to query
    @type  dev_name: C{str}

    @return: ioctl system call return value
    @rtype:  C{str}
    """
    if_struct = struct.pack("256s", dev_name[:15])
    net_val = fcntl.ioctl(sock_fd, ioctl_num, if_struct)
    if ioctl_num == 0x8927:
        hwaddr = ""
        for hw_char in net_val[18:24]:
            hex_num = "00" + hex(ord(hw_char))[2:]
            hwaddr = hwaddr + hex_num[-2:] + ":"
        hwaddr = hwaddr[:len(hwaddr) - 1]
        return hwaddr

    else:
        return socket.inet_ntoa(net_val[20:24])

def _read_port_name(filename):
    f = file(filename, 'r')
    port_wwn = f.readline().rstrip('\n')
    f.close()
    return port_wwn

def getPortNames():
    port_names = []
    for filename in glob('/sys/class/fc_host/*/port_name'):
        port_wwn = _read_port_name(filename)
        port_names.append(port_wwn)
    return port_names

def getInitiatorNames():
    initiators = []
    if os.path.exists('/etc/iscsi/initiatorname.iscsi'):
        f = file('/etc/iscsi/initiatorname.iscsi', 'r')
        while True:
            line = f.readline()
            if line == '':
                break
            line = line.strip()
            if (line == '') or line.startswith('#') or (line.find('=') == -1):
                continue
            (key, value) = line.split('=', 1)
            if (key == 'InitiatorName') or (key == 'InitiatorAlias'):
                initiators.append(value)
        f.close()
    return initiators


def makeMPPage83FromRawPage83(page_83, vendor = None, model = None):
    """
    This will return the multipath uuid from the supplied raw page 83 id.

    @param page_83: The raw page 83 to convert.
    @type  page_83: C{int}

    @param vendor: The lun vendor.
    @type  vendor: C{str}

    @param model: The lun model. 
    @type  model: C{str}
    """

    MAX_SERIAL_LEN = 256
    VENDOR_LENGTH = 8
    MODEL_LENGTH = 16
    PAGE_83 = 0x83

    # id type values of id descriptors. These are assumed to fit in 4 bits.

    SCSI_ID_VENDOR_SPECIFIC = 0
    SCSI_ID_T10_VENDOR = 1
    SCSI_ID_EUI_64 = 2
    SCSI_ID_NAA = 3

    # Supported NAA values. These fit in 4 bits, so the "don't care" value
    # cannot conflict with real values.

    SCSI_ID_NAA_DONT_CARE = 0xff
    SCSI_ID_NAA_IEEE_REG = 5
    SCSI_ID_NAA_IEEE_REG_EXTENDED = 6

    # Supported Code Set values.

    SCSI_ID_BINARY = 1
    SCSI_ID_ASCII = 2

    id_search_list = [
        { 'id_type':SCSI_ID_NAA, 'naa_type':SCSI_ID_NAA_IEEE_REG_EXTENDED, 'code_set':SCSI_ID_BINARY },
        { 'id_type':SCSI_ID_NAA, 'naa_type':SCSI_ID_NAA_IEEE_REG_EXTENDED, 'code_set':SCSI_ID_ASCII },
        { 'id_type':SCSI_ID_NAA, 'naa_type':SCSI_ID_NAA_IEEE_REG, 'code_set':SCSI_ID_BINARY },
        { 'id_type':SCSI_ID_NAA, 'naa_type':SCSI_ID_NAA_IEEE_REG, 'code_set':SCSI_ID_ASCII },
        { 'id_type':SCSI_ID_NAA, 'naa_type':SCSI_ID_NAA_DONT_CARE, 'code_set':SCSI_ID_BINARY },
        { 'id_type':SCSI_ID_NAA, 'naa_type':SCSI_ID_NAA_DONT_CARE, 'code_set':SCSI_ID_ASCII },
        { 'id_type':SCSI_ID_EUI_64,	'naa_type':SCSI_ID_NAA_DONT_CARE, 'code_set':SCSI_ID_BINARY },
        { 'id_type':SCSI_ID_EUI_64,	'naa_type':SCSI_ID_NAA_DONT_CARE, 'code_set':SCSI_ID_ASCII },
        { 'id_type':SCSI_ID_T10_VENDOR,	'naa_type':SCSI_ID_NAA_DONT_CARE, 'code_set':SCSI_ID_BINARY },
        { 'id_type':SCSI_ID_T10_VENDOR,	'naa_type':SCSI_ID_NAA_DONT_CARE, 'code_set':SCSI_ID_ASCII },
        { 'id_type':SCSI_ID_VENDOR_SPECIFIC, 'naa_type':SCSI_ID_NAA_DONT_CARE, 'code_set':SCSI_ID_BINARY },
        { 'id_type':SCSI_ID_VENDOR_SPECIFIC, 'naa_type':SCSI_ID_NAA_DONT_CARE, 'code_set':SCSI_ID_ASCII }]

    serial = ''
    
    if page_83[1] != PAGE_83:
		return None

    if vendor and len(vendor) > VENDOR_LENGTH - 1:
        vendor = vendor[:VENDOR_LENGTH - 1]
    if model and len(model) > MODEL_LENGTH - 1:
        model = model[:MODEL_LENGTH - 1]

    if page_83[6] != 0:   # Extract the raw binary from VPD 0x83 pre-SPC devices 
        serial += "%x" % SCSI_ID_NAA
        for i in range(page_83[3]):
            serial += "%x" % ((page_83[i+4] & 0xf0) >> 4)
            serial += serial + "%x" % (page_83[i+4] & 0x0f)
        return serial

    j = 0

    for entry in id_search_list: 

        # Examine each descriptor returned. There is normally only
        #one or a small number of descriptors.
        
        for j in range(4, page_83[3] + 4, page_83[j + 3] + 4):

            # ASSOCIATION must be with the device (value 0)

            if page_83[j + 1] & 0x30 != 0:
                continue

            if (page_83[j + 1] & 0x0f) != entry['id_type']:
                continue
            
            # Possibly check NAA sub-type. 

            if entry['naa_type'] != SCSI_ID_NAA_DONT_CARE \
                and entry['naa_type'] != ((page_83[j + 4] & 0xf0) >> 4):
                continue

            #Check for matching code set - ASCII or BINARY.

            if (page_83[j] & 0x0f) != entry['code_set']:
                continue

            # page_83[3]: page length

            length = page_83[j + 3]
            
            if (page_83[j] & 0x0f) != SCSI_ID_ASCII:		
                length *= 2    # If not ASCII, use two bytes for each binary value.

            # Add one byte for the NUL termination, and one for the id_type.

            length += 2          

            if entry['id_type'] == SCSI_ID_VENDOR_SPECIFIC:
                length += VENDOR_LENGTH + MODEL_LENGTH

            if MAX_SERIAL_LEN < length:
                continue


            # For SCSI_ID_VENDOR_SPECIFIC prepend the vendor and model before
            # the id since it is not unique across all vendors and models,
            # this differs from SCSI_ID_T10_VENDOR, where the vendor is
            # included in the identifier.

            if entry['id_type'] == SCSI_ID_VENDOR_SPECIFIC:
                serial = vendor + '_' + model + serial

            i = j + 4  # offset to the start of the identifier 
            
            if (page_83[j] & 0x0f) == SCSI_ID_ASCII:
                 # ASCII descriptor.
                temp = ''
                for descriptor_len in range(page_83[j + 3]):
                    temp += chr(page_83[i])
                    i += 1
              
                if temp and temp.startswith(''):
                    temp = temp.lstrip()
                    temp = '_' + temp
                    serial += temp
            else:
                 # Binary descriptor, convert to ASCII, using two bytes of
                 # ASCII for each byte in the page_83.
                serial += "%x" % (entry['id_type'])
                for descriptor_len in range(page_83[j + 3]):
                    serial += "%x" % ((page_83[i] & 0xf0) >> 4)
                    serial += "%x" % (page_83[i] & 0x0f)
                    i += 1

            serial = serial.replace(' ', '_')
            serial = serial.rstrip()
            return serial   

    return None





