#!/usr/bin/env python

"""
@copyright: Copyright (C) 2008, 2010 Oracle and/or its affiliates. All rights reserved.

@license:   See the file COPYING for redistribution information.

@summary:   Simple plugin manager used to discover and use Storage Connect plugins.
"""

import os
import sys
import inspect
import gettext

import OSCPlugin

from OSCPlugin import IPlugin

_ = gettext.gettext

__docformat__ = "epytext en"

class PluginProxy(object):
    """
    Simple proxy class to encapsulate a Storage Connect plugin.
    """

    def __init__(self, plugin_class):
        """
        Proxy class constructor.

        @param plugin_class: Specific Storage Connect plugin class.
        @type  plugin_class: L{IPlugin class<IPlugin.IPlugin>}
        """
        if not issubclass(plugin_class, IPlugin):
            raise TypeError("Supplied class is not a IPlugin class!")

        self.__plugin_class__ = plugin_class
        self.__plugin_methods__ = inspect.getmembers(plugin_class, inspect.ismethod)
        self.__plugin_method_names__ = []
        for name, value in self.__plugin_methods__:
            if not name.startswith("_"):
                self.__setattr__(name, value)
                self.__plugin_method_names__.append(name)

        self.__plugin_functions__ = inspect.getmembers(plugin_class, inspect.isfunction)
        self.__plugin_func_names__ = []
        for name, value in self.__plugin_functions__:
            if not name.startswith("_"):
                self.__setattr__(name, value)
                self.__plugin_func_names__.append(name)

        self.__plugin_info__ = plugin_class.getPluginInfo()
        for name in self.__plugin_info__:
            if not name.startswith("_"):
                self.__setattr__(name, self.__plugin_info__[name])

    def __repr__(self):
        return self.plugin_name                                                     #pylint: disable=E1101

    def __str__(self):
        return "%s version %s by %s - %s" % (self.plugin_name, self.plugin_version, #pylint: disable=E1101
                                             self.vendor_name, self.plugin_desc)    #pylint: disable=E1101

    def __get_functions__(self):
        return self.__plugin_func_names__[:]

    def __get_methods__(self):
        return self.__plugin_method_names__[:]

    def __get_dict__(self):
        return self.__plugin_class__.__dict__

class PluginManager(object):
    """
    Class used by Storage Connect Plugin consumers to discover and use the installed plugins.
    """

    def __init__(self, plugin_logger = None):
        """
        Constructor for the Storage Connect Plugin Manager.
        """
        if not os.path.exists(IPlugin.__plugin_top_dir__):
            raise PluginManagerError("No plugins installed!")

        self.__plugins__ = []
        if IPlugin.__plugin_top_dir__ not in sys.path:
            sys.path.append(IPlugin.__plugin_top_dir__)

        if plugin_logger:
            OSCPlugin.IPlugin.logger = plugin_logger

    def discoverPlugins(self, plugin_types = IPlugin.__plugin_types__, ignore_failed = False):
        """
        Discover all installed Storage Connect plugins on the system.

        @param plugin_types: L{Optional Plugin types to discover<IPlugin.__plugin_types__>}
        @type  plugin_types: C{list}

        @param ignore_failed: Optional flag to ignore all failures
        @type  ignore_failed: C{bool}

        @return: list of newly discovered Storage Connect Plugins.
        @rtype:  C{list}
        """
        failed = []
        plugins = []
        for vendor_dir in os.listdir(IPlugin.__plugin_top_dir__):
            vendor_path = os.path.join(IPlugin.__plugin_top_dir__, vendor_dir)
            if not os.path.isdir(vendor_path) or vendor_dir.startswith('.'):
                continue

            for plugin_mod_dir in os.listdir(vendor_path):
                plugin_mod_path = os.path.join(vendor_path, plugin_mod_dir)
                if os.path.isdir(plugin_mod_path) and not plugin_mod_dir.startswith('.'):
                    plugin_mod_name = "%s.%s" % (vendor_dir, plugin_mod_dir)
                    try:
                        plugin_mod = __import__(plugin_mod_name, globals(), locals(), plugin_mod_dir)

                    except Exception:
                        failed.append(plugin_mod_name)
                        continue

                    for plugin_type in plugin_types:
                        plugin_names = plugin_mod.__dict__.get("osc_%s_plugins" % plugin_type, None)
                        if plugin_names:
                            for plugin_name in plugin_names:
                                if plugin_name not in self.__plugins__:
                                    self.__plugins__.append(Plugin(plugin_name,
                                                                   plugin_mod))
                                    plugins.append(plugin_name)

        if len(failed) > 0 and not ignore_failed:
            raise PluginManagerError(_("Failed to to load plugins: %s" % (','.join(failed))))

        return plugins

    def getPluginManagerClass(self, plugin_name, auto_import = True):
        """
        Get the Plugin Manager class for the named plugin.

        @return: Plugin Manager class for the named Storage Connect plugin.
        @rtype:  C{Plugin}
        """
        for plugin in self.__plugins__:
            if plugin.getName() == plugin_name:
                return plugin

        try:
            plugin_mod_name, class_name = plugin_name.rsplit('.', 1)

        except ValueError:
            raise PluginManagerError(_("Illegal plugin implementation name format,"
                                       " should be vendor.module.file.class"))

        plugin_mod = __import__(plugin_mod_name, globals(), locals(), class_name)
        pkg_name = plugin_mod.__dict__.get("__package__", None)
        if pkg_name:
            vendor_name, mod_name = pkg_name.rsplit('.', 1)
            try:
                plugin_mod = __import__(pkg_name, globals(), locals(), mod_name)

            except ImportError:
                raise PluginManagerError(_("Failed to to load plugin"))

        plugin = Plugin(plugin_name, plugin_mod)
        if auto_import:
            plugin.getClass(auto_import)

        if plugin not in self.__plugins__:
            self.__plugins__.append(plugin)

        return plugin

    def getPlugin(self, plugin_name):
        """
        Get the Plugin class for the named plugin.

        @return: Plugin class for the named Storage Connect plugin.
        @rtype:  C{IPlugin class}
        """
        for plugin in self.__plugins__:
            if plugin.getName() == plugin_name:
                return plugin.getClass()
            
        else:
            plugin = self.getPluginManagerClass(plugin_name)
            return plugin.getClass() 

    def getPluginNames(self):
        """
        Get the list of known plugin names.

        @return: list of known Storage Connect plugin names.
        @rtype:  C{list}
        """
        return [plugin.getName() for plugin in self.__plugins__]

    def getPluginPath(self, plugin_name):
        """
        Get the install path for a specific plugin.

        @param plugin_name: Name of the specific plugin.
        @type  plugin_name: C{str}

        @return: list of newly discovered Storage Connect Plugins.
        @rtype:  C{list}
        """
        for plugin in self.__plugins__:
            if plugin.getName() == plugin_name:
                return plugin.getPath()

        return None

    def getPluginModule(self, plugin_name):
        """
        Get the plugin's module name.

        @param plugin_name: Name of the specific plugin.
        @type  plugin_name: C{str}

        @return: module of the Storage Connect Plugin.
        @rtype:  C{str}
        """
        for plugin in self.__plugins__:
            if plugin.getName() == plugin_name:
                return plugin.getModuleName()

        return None

    def importPlugin(self, plugin_name):
        """
        Import the plugin.

        @param plugin_name: Name of the plugin that should be imported.
        @type  plugin_name: C{str}

        @return: Class object for the plugin.
        @rtype:  C{list}
        """
        for plugin in self.__plugins__:
            if plugin_name == plugin.getName():
                try:
                    return plugin.importPlugin()

                except:
                    if plugin in self.__plugins__:
                        self.__plugins__.remove(plugin)

                    raise

        return None

    def isPluginImported(self, plugin_name):
        """
        Get a list of names for all the imported plugins.

        @param plugin_name: Name of the plugin to check.
        @type  plugin_name: C{str}

        @return: True if the plugin is found and imported.
        @rtype:  C{bool}
        """
        for plugin in self.__plugins__:
            if plugin_name == plugin.getName():
                return plugin.isImported()

        return False

    def getImportedPluginNames(self):
        """
        Get a list of names for all the imported plugins.

        @return: list of names of all the imported plugins.
        @rtype:  C{list}
        """
        return [plugin.getName() for plugin in self.__plugins__ if plugin.isImported()]

    def getPluginInfo(self, plugin_name):
        """
        Get the information for the specified plugin.

        @return: information on and about the Storage Connect plug-in.
        @rtype:  C{dict}
        """
        for plugin in self.__plugins__:
            if plugin.getName() == plugin_name:
                return plugin.getInfo()

        return None

    def getPluginProxy(self, plugin_name):
        """
        Get a proxy class for the plugin.

        @param plugin_name: Name of the plugin to obtain a proxy for.
        @type  plugin_name: C{str}

        @return: The L{Plugin Proxy class<PluginProxy>} for the plugin.
        @rtype:  L{PluginProxy}
        """
        for plugin in self.__plugins__:
            if plugin.getName() == plugin_name:
                return plugin.getProxy()

        return None

    def setLogger(self, plugin_logger):
        """
        Change the logger the plugin should use.

        @param plugin_logger: Set the C{Logger} to use by the plugin
        @type  plugin_logger: C{logger}

        """
        OSCPlugin.IPlugin.logger = plugin_logger

    def __iter__(self):
        """
        Get an iterator object.

        @return: Iterator object.
        @rtype:  C{iter}
        """
        self.__plugins_iter__ = self.__plugins__.__iter__()
        return self

    def next(self):
        """
        Get the next iterator value.

        @return: Next iterator object.
        @rtype:  C{iter}
        """
        return self.__plugins_iter__.next()

class Plugin(object):
    """
    Class encapsulating the Storage Connect plugin
    """

    def __init__(self, plugin_name, plugin_module):
        """
        Construstor for the Plugin encapsulation class.

        @param plugin_name: Name of the plugin.
        @type  plugin_name: C{str}

        @param plugin_module: Plugin module.
        @type  plugin_module: C{object}
        """
        self.__plugin_name__   = plugin_name
        self.__plugin_module__ = plugin_module
        self.__class_module__  = None
        self.__plugin_class__  = None

    def getName(self):
        """
        Get the plugin's fully qualified name.

        @return: Fully Qualified name of the plugin.
        @rtype:  C{str}
        """
        return self.__plugin_name__

    def getPath(self):
        """
        Get the plugin's path.

        @return: Path to the plugin.
        @rtype:  C{str}
        """
        return self.__plugin_module__.__path__[0]

    def getModuleName(self):
        """
        Get the plugin's module name.

        @return: Module name of the plugin.
        @rtype:  C{str}
        """
        return self.__plugin_module__.__name__

    def getClass(self, auto_import = True):
        """
        Get the plugin's list of classes.

        @return: Class objects for the plugin (if it is already imported).
        @rtype:  C{list}
        """
        if not self.__plugin_class__ and auto_import:
            self.importPlugin()

        return self.__plugin_class__

    def importPlugin(self):
        """
        Import the IPlugin class.
        """
        mod_name, class_name = self.__plugin_name__.rsplit('.', 1)
        try:
            self.__class_module__ = __import__(mod_name, globals(), locals(), class_name)

        except ImportError, imp_err:
            if imp_err.__dict__.has_key("args"):
                raise PluginManagerError(_("Unable to import plugin %s (%s)" % (self.__plugin_name__, imp_err.args[0])))

            else:    
                raise PluginManagerError(_("Unable to import plugin %s (%s)" % (self.__plugin_name__, imp_err.message)))

        self.__plugin_class__ = self.__class_module__.__dict__.get(class_name, None)
        if not self.__plugin_class__:
            raise PluginManagerError(_("Unable to import plugin %s" % self.__plugin_name__))
        
        required_api_vers = self.__plugin_class__.__dict__.get("required_api_vers", None)
        if required_api_vers is None:
            raise PluginManagerError(_("Unable to import plugin %s, plugin did not set minimum API level!" % self.__plugin_name__))
        
        if int("".join(required_api_vers)) > int("".join(IPlugin.plugin_api_version)):
            raise PluginManagerError(_("Unable to import plugin %s, plugin require API version %.2f" %
                                       (self.__plugin_name__,
                                        int("".join(required_api_vers)) / 100.0)))

    def isImported(self):
        """
        Check if the Plugin is already imported.

        @return: True if the Plugin is already imported.
        @rtype:  C{bool}
        """
        return (self.__plugin_class__ != None)

    def getInfo(self):
        """
        Get the information for the plugin.

        @return: information on and about the Storage Connect plug-in.
        @rtype:  C{dict}
        """
        p_class = self.getClass(auto_import = True)
        p_info = p_class.getPluginInfo()
        return p_info

    def getProxy(self):
        """
        Get a proxy instance for the plugin class.

        @return: Plugin proxy instance.
        @rtype:  L{Plugin proxy instance<PluginProxy>}
        """
        if self.__plugin_class__:
            return PluginProxy(self.__plugin_class__)

        return None

    def __repr__(self):
        """
        Allow displaying of the plugin names.

        @return: Plugin name.
        @rtype:  C{str}
        """
        return self.__plugin_name__

class PluginManagerError(Exception):
    """
    Plugin manager error exception.
    """
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)
