#!/usr/bin/env python

"""
@copyright: Copyright (C) 2008, 2010 Oracle and/or its affiliates. All rights reserved.

@license:   See the file COPYING for redistribution information.

@summary:   Simulate the classic C enum type (with some Pythonic improvements).
"""

import re

class EnumUpdateError(Exception):
    """
    Exception raised when trying update any PyEnum attributes directly.
    """
    def __init__(self, message):
        """
        @param message: Error message for the exception.
        @type  message: C{str}
        """
        self.message = message

class PyEnum(object):
    """
    Class that simulate the C enum type

    @note: PyEnum does not allow general access to the class variables
    """
    def __init__(self, *enum_ids):
        """
        Create the enum instance.

        @param enum_ids: List of the enum identifiers, an optional value
                         can be associated with the key in the form of
                         name=value
        @type  enum_ids: C{list}
        """
        enum_value = 0
        enum_re = re.compile("\s*[=]\s*")
        self.__enum_ids__ = {}
        for enum_id in enum_ids:
            enum_pair = enum_re.split(enum_id.strip())
            if len(enum_pair) < 1 or len(enum_pair) > 2:
                raise SyntaxError("invalid enum value syntax")

            elif len(enum_pair) == 2:
                enum_value = int(enum_pair[1], 0)

            self.__enum_ids__[enum_pair[0]] = enum_value
            enum_value += 1

    def __getattr__(self, enum_id):
        """
        Get the value for a specific enum id.

        @param enum_id: Identifier of the enum value to retrieve.
        @type  enum_id: C{str}

        @return: The enum value.
        @rtype:  C{int}

        @raise AttributeError: on failure
        """
        try:
            return self.__enum_ids__[enum_id]

        except:
            raise AttributeError

    def __getitem__(self, enum_id):
        """
        Get the value for a specific enum id.

        @param enum_id: Identifier of the enum value to retrieve
        @type  enum_id: C{str}

        @return: The enum value.
        @rtype:  C{int}

        @raise AttributeError: on failure
        """
        if type(enum_id) == str:
            return self.__getattr__(enum_id)
        else:
            for enum_key in self.__enum_ids__:
                if self.__enum_ids__[enum_key] == enum_id:
                    return enum_key

    def __setattr__(self, attr_name, attr_val):
        """
        Set the value for a specific enum id.

        @param attr_name: Identifier of the attribute to set
        @type  attr_name: C{str}

        @param attr_val: Object to assing to the attribute
        @type  attr_val: C{object}

        @note: PyEnum does not allow updating the enums, it will raise
               EnumUpdateError.

        @raise EnumUpdateError: on failure
        """
        if attr_name == "__enum_ids__" or \
           attr_name == "__enum_iter__":
            self.__dict__[attr_name] = attr_val

        else:
            raise EnumUpdateError("No changes to %s allowed" % attr_name)

    def __iter__(self):
        """
        Get an iterator object.

        @return: Iterator object.
        @rtype:  C{iter}
        """
        self.__enum_iter__ = self.__enum_ids__.__iter__()
        return self

    def next(self):
        """
        Get the next iterator value.

        @return: Next iterator object.
        @rtype:  C{iter}
        """
        return self.__enum_iter__.next()

    def getID(self, enum_value):
        """
        Get the the id for a specific value.
        @note: The ID of the first matching value will be returned.

        @param enum_value: Enum value to get the identifier for.
        @type  enum_value: C{str}

        @return: The enum identifier.
        @rtype:  C{str}

        @raise AttributeError: on failure
        """
        try:
            for enum_key in self.__enum_ids__:
                if self.__enum_ids__[enum_key] == enum_value:
                    return enum_key

        except:
            raise AttributeError

    def dump(self):
        """
        Dump all the enum identifier and values.
        """
        for enum_id in self.__enum_ids__.keys():
            print "%s=%s" % (enum_id, self.__enum_ids__[enum_id])

    def toString(self):
        """
        Return all the enum identifier and values in a comma delimited string.
        """
        enum_str = ""
        for enum_id in self.__enum_ids__.keys():
            enum_str = enum_str + "%s=%s," % (enum_id, 
                                              self.__enum_ids__[enum_id])

        return enum_str.rstrip(',')
